/* patches */
.diffChunk div.section {
 width: auto;
}

div.diffComment {
  margin-bottom: 1.8em;
  color: #333;
}

span.diffChunkTitle {
 font-weight: bold;
}

div.diffCmd {
 margin: 1em 0 0;
 padding: 4px;
 border: 1px #AAA solid;
 background-color: #C2C2C2;
 font-weight: bold;
 font-size: 130%;
}

div.diffChunk pre {
 margin: 0.5em 0;
 padding: 0 0.3em;
 border-color: #dfdfdf;
 line-height: 1.2m;
}

div.diffChunk pre span.diffOldLine {
 color:#FF0000;
}

div.diffChunk pre span.diffNewLine {
  color: green;
}

pre.diff {
  font-size: 110%;
}
